/*	File:	init.c
	Author:	Alex Fedosov, Jeff Malch
	Date:	December 13, 2000
	Description:	common init functions
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common/types.h"
#include "common/const.h"
#include "common/errors.h"
#include "common/map.h"
#include "common/object.h"
#include "common/init.h"
#include "common/util.h"

#include "dmzcore.h"
#include "dmzgraph.h"

#define YELLOW 0xC0C000 //16776960
#define GREEN 0xC000 //65280
#define BLUE 0x80 //255
#define CYAN 0xC0C0

#define BASE_W2G 25
#define BASE_S2G 37
#define BASE_S2W 49
#define BASE_W2A 61

/* this function is defined in the client, and stubbed in server */
void DrawProgressIndicator(int percent, int flag);

char buf[256];

int dmz_data_init(void)
{
	int error;
	
	srand(DMZ_RANDOM_SEED);

	dmz_map_init();
	dmz_object_init();

	DrawProgressIndicator(0, 0);

	dmz_logstd("[common:init:dmz_data_init] building terrain");
	error = dmz_load_terrain(TERRAIN_CSV, MAP_FILE);
	if (error != DMZ_SUCCESS) return error;

	DrawProgressIndicator(20, 0);

	/* read in CSV files */
	dmz_logstd("[common:init:dmz_data_init] loading objects");
	error = dmz_load_CSV(OBJECTS_CSV);
	if (error != DMZ_SUCCESS) return error;

	DrawProgressIndicator(40, 0);

	dmz_logstd("[common:init:dmz_data_init] loading monsters");
	error = dmz_load_CSV(MONSTERS_CSV);
	if (error != DMZ_SUCCESS) return error;

	DrawProgressIndicator(60, 0);

	dmz_logstd("[common:init:dmz_data_init] loading items");
	error = dmz_load_CSV(ITEMS_CSV);
	if (error != DMZ_SUCCESS) return error;

	DrawProgressIndicator(70, 0);

/*
	dmz_logstd("[common:init:dmz_data_init] loading NPCs");
	error = dmz_load_CSV(NPC_CSV);
	if (error != DMZ_SUCCESS) return error;
*/
	dmz_logstd("[common:init:dmz_data_init] done");

	return DMZ_SUCCESS;
}

int dmz_load_CSV(char *filename)
{
	FILE *fp;
	char csvbuffer[MAXLINE], *rv;
	char *csvtext[CSV_FIELDS];
	int i;
	
	fp = fopen(filename, "r");
	if (fp == NULL)
		return DMZ_DATA_ERROR_OBJECT;

	while (1)
	{
		WORD16 obj, type;
		short csvfields[CSV_FIELDS];

		rv = fgets(csvbuffer, MAXLINE, fp);
		if (rv == NULL) break;

		i = 0;

		rv = strtok(csvbuffer, ",");

		/* csv format for objects, monsters, items:
			Name,Alive Bitmap,Dead Bitmap,Hitpoints,Damage,Armor,Value,Kind,X,Y

			Kind applies only to items, is an ITEMTYPE enum, and can be one of:
				1 -- armor upgrade
				2 -- weapon upgrade
				3 -- ammo recharge
				4 -- speed upgrade

			for non-items, kind should be set to 0 or negative value
		*/

		while (rv != NULL)
		{
			csvtext[i++] = rv;
			rv = strtok(NULL, ",");
		}

		if (i != CSV_FIELDS) return DMZ_DATA_ERROR_OBJECT;

		/* convert csv text into something meaningful */
		for (i = CSV_HP; i <= CSV_Y; i++)
			csvfields[i] = atoi(csvtext[i]);

		/* determine type of object */
		if (csvfields[CSV_KIND] > 0)
			type = ITEM;
		else if (csvfields[CSV_ARM] == 0 && csvfields[CSV_DAM] == 0)
			if (csvfields[CSV_VAL] == 0)
				type = NPC;
			else
				type = TERRAIN;
		else
			type = MONSTER;

		if (csvfields[CSV_X] == 0) csvfields[CSV_X] = rand() % map_width;
		if (csvfields[CSV_Y] == 0) csvfields[CSV_Y] = rand() % map_height;

		obj = dmz_new_object(0, type, csvfields[CSV_X] * PIXELS_PER_CELL, csvfields[CSV_Y] * PIXELS_PER_CELL);

		/* now attempt to create an object */
		csvfields[CSV_BIT] = dmz_graph_load_image(csvtext[CSV_BIT],
			&dmz_objects[obj]->width, &dmz_objects[obj]->height);

		csvfields[CSV_DEAD] = dmz_graph_load_image(csvtext[CSV_DEAD],
			NULL, NULL);

		if (csvfields[CSV_BIT] <= 0 || csvfields[CSV_DEAD] <= 0)
		{
			dmz_logerr("[common:init:dmz_load_csv] error loading bitmap");
			dmz_logstd(csvtext[CSV_BIT]);
			return DMZ_DATA_ERROR_BITMAP;
		}

		/* fill in name, damage, armor, value */
		strtoupr(csvtext[CSV_NAME]);
		strcpy(dmz_objects[obj]->name, csvtext[CSV_NAME]);
		dmz_objects[obj]->bitidx = csvfields[CSV_BIT];
		dmz_objects[obj]->deadbit = csvfields[CSV_DEAD];
		dmz_objects[obj]->score = csvfields[CSV_VAL];

		switch (type)
		{
			case NPC:
				dmz_objects[obj]->width = DEFAULT_WIDTH;
				dmz_objects[obj]->height = DEFAULT_HEIGHT;
				break;
			case MONSTER:
				dmz_objects[obj]->u.player.hth_dam = csvfields[CSV_DAM];
				dmz_objects[obj]->u.player.life = csvfields[CSV_HP];
				dmz_objects[obj]->width = DEFAULT_WIDTH;
				dmz_objects[obj]->height = DEFAULT_HEIGHT;
				break;
			case TERRAIN:
				dmz_objects[obj]->u.terrain.ter_hp = csvfields[CSV_HP];
				break;
			case ITEM:
				dmz_objects[obj]->u.item.item_dam = csvfields[CSV_DAM];
				dmz_objects[obj]->u.item.item_arm = csvfields[CSV_ARM];
				dmz_objects[obj]->width = DEFAULT_WIDTH;
				dmz_objects[obj]->height = DEFAULT_HEIGHT;

				if (csvfields[CSV_KIND] == AMMOUPGRADE)
				{
					dmz_objects[obj]->u.item.item_ammo = csvfields[CSV_VAL];
					dmz_objects[obj]->u.item.kind = AMMOUPGRADE;
				}

				else
				{
					dmz_objects[obj]->u.item.item_ammo = 0;
					dmz_objects[obj]->u.item.kind = csvfields[CSV_KIND];
				}

				break;
			default:
				return DMZ_DATA_ERROR;
		}

		sprintf(buf, "adding object %s type %d bitmap %d", dmz_objects[obj]->name, dmz_objects[obj]->objtype, dmz_objects[obj]->bitidx);
		dmz_logstd(buf);
	}

	return DMZ_SUCCESS;
}

int dmz_load_terrain(char *csvfile, char *mapfile)
{
	FILE *fp;
	char csvbuffer[MAXLINE], *rv;
	char *csvtext[CSV_TER_FIELDS];
	unsigned char *newmap = NULL;
	unsigned int off, color, tile, i;
	unsigned int left = 0, top = 0, diag = 0;
	int bmw, bmh;

	dmz_logstd("[common:init:dmz_data_init] loading map");
	newmap = dmz_graph_load_bitmap(mapfile, &bmw, &bmh);
	if (newmap == NULL)	return DMZ_DATA_ERROR_MAP;

	dmz_logstd("[common:init:dmz_data_init] loading terrain");
	fp = fopen(csvfile, "r");
	if (fp == NULL)
		return DMZ_DATA_ERROR_TERRAIN;

	init_groups();

	while (1)
	{
		int csvfields[CSV_TER_FIELDS];
		CELL c;

		rv = fgets(csvbuffer, MAXLINE, fp);
		if (rv == NULL) break;

		i = 0;

		rv = strtok(csvbuffer, ",");

		/* csv format for terrain:
			File Name,Group,Probability,Passability,Speed,Damage
		*/

		while (rv != NULL)
		{
			csvtext[i++] = rv;
			rv = strtok(NULL, ",");
		}

		if (i != CSV_TER_FIELDS) return DMZ_DATA_ERROR_TERRAIN;

		/* convert csv text into something meaningful */
		for (i = CSV_TER_GRP; i <= CSV_TER_DAM; i++)
			csvfields[i] = atoi(csvtext[i]);

		/* now attempt to create a map cell */
		csvfields[CSV_TER_FILE] = dmz_graph_load_image(csvtext[CSV_TER_FILE],
			NULL, NULL);

		if (csvfields[CSV_TER_FILE] < 0)
		{
			sprintf(buf, "load_image() returned %d", csvfields[CSV_TER_FILE]);
			dmz_logerr(buf);
			return DMZ_DATA_ERROR_BITMAP;
		}

		/* create a new map cell */
		c = dmz_map_new_cell((WORD16) csvfields[CSV_TER_FILE]);

		dmz_map_set_prop(c, (short *) &csvfields[CSV_TER_PASS], (short *) &csvfields[CSV_TER_DAM], (short *) &csvfields[CSV_TER_SPD]);

		/* also give tile to randomizer */
		if (csvfields[CSV_TER_GRP] != 0)
		{
			if (!add_to_group(csvfields[CSV_TER_FILE], csvfields[CSV_TER_GRP], csvfields[CSV_TER_PROB]))
				return DMZ_DATA_ERROR_TERRAIN;
		}

/*		sprintf(buf, "adding terrain %d %s", c, csvtext[CSV_TER_FILE]);
		dmz_logstd(buf);
*/	}

	map_width = bmw;
	if (bmh < 0) map_height = 0 - bmh;
	else map_height = bmh;

	off = 0;

	dmz_logstd("[common:init:dmz_load_resource] going through bitmap");
	for (i = 0; i < map_width * map_height; i++)
	{
		unsigned int x, y, z;
		int rv;
		y = (off / 3) / map_width;
		x = (off / 3) % map_width;

		color = (newmap[off + 2] << 16) | (newmap[off + 1] << 8) | newmap[off];
		off += 3;

		y = map_width - 1 - y;
		z = y * map_width + x;

		tile = get_random_tile(color);

		/* use transition tile if applicable */
		rv = find_transition_tile(color, left, top, diag);
		if (rv != 0) tile = rv;

		dmz_world_map[x][y] = dmz_map_new_cell((WORD16) tile);

		left = color;
		if (y > 0)
		{
			int t_off = off + map_width * 3;
			top = (newmap[t_off + 2] << 16) | (newmap[t_off + 1] << 8) | newmap[t_off];
			if (x > 0)
			diag = (newmap[t_off - 1] << 16) | (newmap[t_off - 2] << 8) | newmap[t_off - 3];
		}
	}

	dmz_logstd("[common:init:dmz_load_resource] done");

	destroy_groups();

//	free(newmap);

	return DMZ_SUCCESS;
}

int find_transition_tile(int cur, int left, int top, int diag)
{
	int rv = 0;

	/* grass - desert transition */
		 if (cur == GREEN && left == GREEN && top == GREEN && diag == GREEN)
		rv = get_random_tile(cur);
	else if (cur == GREEN && left == GREEN && top == GREEN && diag == YELLOW)
		rv = 11 + BASE_S2G;
	else if (cur == GREEN && left == GREEN && top == YELLOW && diag == GREEN)
		rv = 8 + BASE_S2G;
	else if (cur == GREEN && left == GREEN && top == YELLOW && diag == YELLOW)
		rv = 3 + BASE_S2G;
	else if (cur == GREEN && left == YELLOW && top == GREEN && diag == GREEN)
		rv = 10 + BASE_S2G;
	else if (cur == GREEN && left == YELLOW && top == GREEN && diag == YELLOW)
		rv = 1 + BASE_S2G;
	else if (cur == GREEN && left == YELLOW && top == YELLOW && diag == GREEN)
		rv = 0;	/* anything here? */
	else if (cur == GREEN && left == YELLOW && top == YELLOW && diag == YELLOW)
		rv = 2 + BASE_S2G;
	else if (cur == YELLOW && left == GREEN && top == GREEN && diag == GREEN)
		rv = 9 + BASE_S2G;
	else if (cur == YELLOW && left == GREEN && top == GREEN && diag == YELLOW)
		rv = 0; /* anything here? */
	else if (cur == YELLOW && left == GREEN && top == YELLOW && diag == GREEN)
		rv = 5 + BASE_S2G;
	else if (cur == YELLOW && left == GREEN && top == YELLOW && diag == YELLOW)
		rv = 4 + BASE_S2G;
	else if (cur == YELLOW && left == YELLOW && top == GREEN && diag == GREEN)
		rv = 7 + BASE_S2G;
	else if (cur == YELLOW && left == YELLOW && top == GREEN && diag == YELLOW)
		rv = 0 + BASE_S2G;
	else if (cur == YELLOW && left == YELLOW && top == YELLOW && diag == GREEN)
		rv = 6 + BASE_S2G;
	else if (cur == YELLOW && left == YELLOW && top == YELLOW && diag == YELLOW)
		rv = get_random_tile(cur);



	/* grass - water transition */
	else if (cur == GREEN && left == GREEN && top == GREEN && diag == GREEN)
		rv = get_random_tile(cur);
	else if (cur == GREEN && left == GREEN && top == GREEN && diag == BLUE)
		rv = 11 + BASE_W2G;
	else if (cur == GREEN && left == GREEN && top == BLUE && diag == GREEN)
		rv = 8 + BASE_W2G;
	else if (cur == GREEN && left == GREEN && top == BLUE && diag == BLUE)
		rv = 3 + BASE_W2G;
	else if (cur == GREEN && left == BLUE && top == GREEN && diag == GREEN)
		rv = 10 + BASE_W2G;
	else if (cur == GREEN && left == BLUE && top == GREEN && diag == BLUE)
		rv = 1 + BASE_W2G;
	else if (cur == GREEN && left == BLUE && top == BLUE && diag == GREEN)
		rv = 0;	/* anything here? */
	else if (cur == GREEN && left == BLUE && top == BLUE && diag == BLUE)
		rv = 2 + BASE_W2G;
	else if (cur == BLUE && left == GREEN && top == GREEN && diag == GREEN)
		rv = 9 + BASE_W2G;
	else if (cur == BLUE && left == GREEN && top == GREEN && diag == BLUE)
		rv = 0; /* anything here? */
	else if (cur == BLUE && left == GREEN && top == BLUE && diag == GREEN)
		rv = 5 + BASE_W2G;
	else if (cur == BLUE && left == GREEN && top == BLUE && diag == BLUE)
		rv = 4 + BASE_W2G;
	else if (cur == BLUE && left == BLUE && top == GREEN && diag == GREEN)
		rv = 7 + BASE_W2G;
	else if (cur == BLUE && left == BLUE && top == GREEN && diag == BLUE)
		rv = 0 + BASE_W2G;
	else if (cur == BLUE && left == BLUE && top == BLUE && diag == GREEN)
		rv = 6 + BASE_W2G;
	else if (cur == BLUE && left == BLUE && top == BLUE && diag == BLUE)
		rv = get_random_tile(cur);


	/* water - desert transition */
	else if (cur == BLUE && left == BLUE && top == BLUE && diag == BLUE)
		rv = get_random_tile(cur);
	else if (cur == BLUE && left == BLUE && top == BLUE && diag == YELLOW)
		rv = 6 + BASE_S2W;
	else if (cur == BLUE && left == BLUE && top == YELLOW && diag == BLUE)
		rv = 0 + BASE_S2W;
	else if (cur == BLUE && left == BLUE && top == YELLOW && diag == YELLOW)
		rv = 7 + BASE_S2W;
	else if (cur == BLUE && left == YELLOW && top == BLUE && diag == BLUE)
		rv = 4 + BASE_S2W;
	else if (cur == BLUE && left == YELLOW && top == BLUE && diag == YELLOW)
		rv = 5 + BASE_S2W;
	else if (cur == BLUE && left == YELLOW && top == YELLOW && diag == BLUE)
		rv = 0;	/* anything here? */
	else if (cur == BLUE && left == YELLOW && top == YELLOW && diag == YELLOW)
		rv = 9 + BASE_S2W;
	else if (cur == YELLOW && left == BLUE && top == BLUE && diag == BLUE)
		rv = 2 + BASE_S2W;
	else if (cur == YELLOW && left == BLUE && top == BLUE && diag == YELLOW)
		rv = 0; /* anything here? */
	else if (cur == YELLOW && left == BLUE && top == YELLOW && diag == BLUE)
		rv = 1 + BASE_S2W;
	else if (cur == YELLOW && left == BLUE && top == YELLOW && diag == YELLOW)
		rv = 10 + BASE_S2W;
	else if (cur == YELLOW && left == YELLOW && top == BLUE && diag == BLUE)
		rv = 3 + BASE_S2W;
	else if (cur == YELLOW && left == YELLOW && top == BLUE && diag == YELLOW)
		rv = 8 + BASE_S2W;
	else if (cur == YELLOW && left == YELLOW && top == YELLOW && diag == BLUE)
		rv = 11 + BASE_S2W;
	else if (cur == YELLOW && left == YELLOW && top == YELLOW && diag == YELLOW)
		rv = get_random_tile(cur);


	/* water - arctic transition */
	else if (cur == BLUE && left == BLUE && top == BLUE && diag == BLUE)
		rv = get_random_tile(cur);
	else if (cur == BLUE && left == BLUE && top == BLUE && diag == CYAN)
		rv = 6 + BASE_W2A;
	else if (cur == BLUE && left == BLUE && top == CYAN && diag == BLUE)
		rv = 0 + BASE_W2A;
	else if (cur == BLUE && left == BLUE && top == CYAN && diag == CYAN)
		rv = 7 + BASE_W2A;
	else if (cur == BLUE && left == CYAN && top == BLUE && diag == BLUE)
		rv = 4 + BASE_W2A;
	else if (cur == BLUE && left == CYAN && top == BLUE && diag == CYAN)
		rv = 5 + BASE_W2A;
	else if (cur == BLUE && left == CYAN && top == CYAN && diag == BLUE)
		rv = 0;	/* anything here? */
	else if (cur == BLUE && left == CYAN && top == CYAN && diag == CYAN)
		rv = 9 + BASE_W2A;
	else if (cur == CYAN && left == BLUE && top == BLUE && diag == BLUE)
		rv = 2 + BASE_W2A;
	else if (cur == CYAN && left == BLUE && top == BLUE && diag == CYAN)
		rv = 0; /* anything here? */
	else if (cur == CYAN && left == BLUE && top == CYAN && diag == BLUE)
		rv = 1 + BASE_W2A;
	else if (cur == CYAN && left == BLUE && top == CYAN && diag == CYAN)
		rv = 10 + BASE_W2A;
	else if (cur == CYAN && left == CYAN && top == BLUE && diag == BLUE)
		rv = 3 + BASE_W2A;
	else if (cur == CYAN && left == CYAN && top == BLUE && diag == CYAN)
		rv = 8 + BASE_W2A;
	else if (cur == CYAN && left == CYAN && top == CYAN && diag == BLUE)
		rv = 11 + BASE_W2A;
	else if (cur == CYAN && left == CYAN && top == CYAN && diag == CYAN)
		rv = get_random_tile(cur);


	return rv;
}

/* below is the support code for generating random terrain */

struct tile_group
{
	unsigned int tiles[MAXTILESPERGROUP];
	unsigned int number;
	unsigned int color;
	unsigned int sum;
	unsigned int prev;
	unsigned int *randomness;
};

struct tile_group tile_groups[MAXGROUPS];

#define FREE(x, s) if (x != NULL) { dmz_free(x, s); x = NULL; }
#define RAND_SIZE 1024 * sizeof(unsigned int)

void init_groups(void)
{
	int i;

	for (i = 0; i < MAXGROUPS; i++)
	{
		tile_groups[i].color = 0;
		tile_groups[i].number = 0;
		tile_groups[i].sum = 0;
		tile_groups[i].prev = 0;
		tile_groups[i].randomness = NULL;
	}
}

void destroy_groups(void)
{
	int i;

	for (i = 0; i < MAXGROUPS; i++)
	{
		FREE(tile_groups[i].randomness, RAND_SIZE);
	}
}

int add_to_group(unsigned int tile, unsigned int grp, unsigned int prob)
{
	unsigned int index = grp % MAXGROUPS;
	unsigned int i, num, pmax;

	if (tile_groups[index].number == 0)			/* group empty */
		tile_groups[index].color = grp;

	else if (grp != tile_groups[index].color)	/* wrong group */
		return 0;

	else if (tile_groups[index].number >= MAXTILESPERGROUP)	/* group full */
		return 0;

	num = tile_groups[index].number;

	for (i = 0; i < num; i++)					/* already in group */
		if (tile_groups[index].tiles[i] == tile)
			return num;

	tile_groups[index].tiles[num] = tile;
	tile_groups[index].sum += prob;

	pmax = prob + tile_groups[index].prev;

	if (tile_groups[index].randomness == NULL)
		tile_groups[index].randomness =
		(unsigned int *) dmz_malloc(RAND_SIZE);

	/* also add to randomness array */
	for (i = tile_groups[index].prev; i < pmax; i++)
		tile_groups[index].randomness[i] = tile;

	tile_groups[index].prev += prob;

	return ++tile_groups[index].number;
}

unsigned int get_random_tile(int grp)
{
	unsigned int index = grp % MAXGROUPS;
	unsigned int tile, r;

	if (tile_groups[index].number == 0) return 0;

	/* now "randomly" select a tile */
	r = rand() % tile_groups[index].sum;
	tile = tile_groups[index].randomness[r];

	return tile;
}
