/*	File:	linklist.c
	Author:	Alex Fedosov
	Date:	February 2, 2001
	Description:	Generic linked list implementation
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sys/linklist.h"

#include "dmzcore.h"

#define FREE(x) if (x != NULL) { free(x); x = NULL; }

void LL_free_list(LIST_NODE *list);

LINK_LIST* LL_init(int numelems, int elemsize)
{
	LINK_LIST *myLL;
	LIST_NODE *e, *tail;
	int i;

	if (numelems < 1 || elemsize < 0) return NULL;

	myLL = (LINK_LIST *) malloc(sizeof(LINK_LIST));

	if (myLL == NULL)
		return NULL;

	myLL->elemsize = elemsize;
	myLL->numelems = numelems;
	myLL->cur_num = 0;

	myLL->head = NULL;
	myLL->tail = NULL;

	/* create a pool of pre-allocated list elements
	   note: list is created backwards */
	tail = NULL;
	for (i = 0; i < numelems; i++)
	{
		/* first allocate the node itself */
		e = (LIST_NODE *) malloc(sizeof(LIST_NODE));
		if (e == NULL)
		{
			free(myLL);
			return NULL;
		}

		if (elemsize > 0)
		{
			e->data = malloc(elemsize);
			if (e->data == NULL)
			{
				free(e);
				free(myLL);
				return NULL;
			}
		}
		else
			e->data = NULL;

		/* point to the logically next, but previously created node */
		e->next = tail;
		tail = e;
	}

	/* we now have a pool of pre-allocated list nodes */
	myLL->pool = tail;

	return myLL;
}

void LL_destroy(LINK_LIST *myLL)
{
	if (myLL == NULL) return;

	/* determine if data needs to be freed */
	LL_free_list(myLL->head);
	LL_free_list(myLL->pool);

	FREE(myLL);
}

int LL_add(LINK_LIST *myLL, void *elem, int reserved)
{
	LIST_NODE *node;

	if (myLL == NULL) return 0;

	/* get a new node from the pool */
	node = myLL->pool;

	if (node == NULL) return 0;

	myLL->pool = node->next;

	/* copy data into list element */
	if (myLL->elemsize > 0)
	{
		if (elem == NULL) return 0;
		memset(node->data, 0, myLL->elemsize);
		memcpy(node->data, elem, myLL->elemsize);
	}

	/* set the user-defined field */
	node->reserved = reserved;
	node->next = NULL;

	/* update tail */
	if (myLL->head == NULL) myLL->head = node;
	if (myLL->tail != NULL) myLL->tail->next = node;
	myLL->tail = node;

	myLL->cur_num++;

	return ~0;
}

int LL_remove(LINK_LIST *myLL, void *dest)
{
	LIST_NODE *node;

	if (myLL == NULL) return 0;

	node = myLL->head;

	if (node == NULL) return 0;

	/* copy data */
	if (myLL->elemsize > 0)
	{
		if (dest == NULL) return 0;
		memcpy(dest, node->data, myLL->elemsize);
	}

	/* remove the node from the list */
	/* need to update tail only if last element removed */
	if (myLL->head == myLL->tail)
		myLL->tail = NULL;

	myLL->head = node->next;

	node->next = myLL->pool;
	myLL->pool = node;

	myLL->cur_num--;

	return node->reserved;
}

void LL_free_list(LIST_NODE *list)
{
	LIST_NODE *current = list, *previous = NULL;

	if (list == NULL) return;

	while (current != NULL)
	{
		FREE(current->data);
		FREE(previous);
		previous = current;
		current = current->next;
	}

	/* free the last element */
	FREE(previous);
}

int LL_num_elements(LINK_LIST *myLL)
{
	if (myLL == NULL)
		return 0;
	else
		return myLL->cur_num;
}
