/*	File:	object.c
	Author:	Alex Fedosov
	Date:	October 19, 2000
	Description:	implementation of game objects
*/

#include <stdio.h>
#include <stdlib.h>

#include "common/types.h"
#include "common/const.h"
#include "common/errors.h"
#include "common/object.h"

#include "dmzcore.h"

#define FREE(p, s) if (p != NULL) { dmz_free(p, s); p = NULL; }

HOBJECT dmz_objects[MAXOBJECTS];

char buf[256];

void dmz_object_init(void)
{
	int i;

	for (i = 0; i < MAXOBJECTS; i++)
		dmz_objects[i] = NULL;
}

void dmz_object_destroy(void)
{
	int i;

	for (i = 0; i < MAXOBJECTS; i++)
		FREE(dmz_objects[i], sizeof(struct dmz_object));
}

WORD16 dmz_object_get_available_id(void)
{
	WORD16 i;

	for (i = 1; i < MAXOBJECTS; i++)
	{
		if (dmz_objects[i] == NULL)
			return i;
	}

	return 0;
}

/* create a new object, given the information about it. if the requested id
   is 0, pick next available id. */
WORD16 dmz_new_object(WORD16 objid, WORD16 objtype, WORD32 init_x, WORD32 init_y)
{
	int i;

	/* try to use requested id */
	if (objid == 0)
	{
		objid = dmz_object_get_available_id();

		if (objid == 0)
		{
			dmz_logerr("[common:object:dmz_get_objid] ran out of object id's!");
			return 0;
		}
	}

	if (dmz_objects[objid] == NULL)
	{
		dmz_objects[objid] = (HOBJECT) dmz_malloc(sizeof(struct dmz_object));

		if (dmz_objects[objid] == NULL)
		{
			dmz_logerr("[common:object:dmz_new_object] unable to allocate memory!");
			return 0;
		}

		memset(dmz_objects[objid], 0, sizeof(struct dmz_object));
	}

	else	/* cell already used? */
		dmz_logerr("[common:object:dmz_new_object] found unexpected object, overwriting\n");

	dmz_objects[objid]->objid = objid;
	dmz_objects[objid]->objtype = objtype;
	dmz_objects[objid]->x = init_x;
	dmz_objects[objid]->y = init_y;

	dmz_objects[objid]->bitidx = 0;
	dmz_objects[objid]->deadbit = 0;
	dmz_objects[objid]->velocity = DEFAULT_VELOCITY;
	dmz_objects[objid]->status = IDLE;
	dmz_objects[objid]->score = 0;
	dmz_objects[objid]->tstamp = 0;
	
	dmz_objects[objid]->width = DEFAULT_WIDTH;
	dmz_objects[objid]->height = DEFAULT_HEIGHT;

	for (i = 0; i < MAX_INV; i++)
		dmz_objects[objid]->inv[i] = 0;

	/* further initialization of object based on type */
	switch (objtype)
	{
	/* player and monster have the same structures */
	case PLAYER:
	case MONSTER:
		dmz_objects[objid]->u.player.life = DEFAULT_LIFE;
		dmz_objects[objid]->u.player.direction = DEFAULT_DIR;
		dmz_objects[objid]->u.player.frame = 0;
		dmz_objects[objid]->u.player.cur_item = 0;
		dmz_objects[objid]->u.player.level = 0;
		dmz_objects[objid]->u.player.frags = 0;
		dmz_objects[objid]->u.player.hth_dam = HTH_DAMAGE;
		dmz_objects[objid]->u.player.mis_dam = MISSILE_DAMAGE;
		dmz_objects[objid]->u.player.ammo = DEFAULT_AMMO;
		break;

	case NPC:
		dmz_objects[objid]->u.npc.npc_dir = DEFAULT_DIR;
		dmz_objects[objid]->u.npc.npc_walk = 0;
		break;

	case ITEM:
		dmz_objects[objid]->u.item.item_dam = 0;
		dmz_objects[objid]->u.item.item_arm = 0;
		break;

	case TERRAIN:
		dmz_objects[objid]->u.terrain.ter_hp = 0;
		break;

	case MISSILE:
		dmz_objects[objid]->velocity = MISSILE_VELOCITY;
		dmz_objects[objid]->status = MOVING;
		dmz_objects[objid]->u.missile.range = 0;
		dmz_objects[objid]->u.missile.owner = 0;
		dmz_objects[objid]->u.missile.direction = NORTH;
		dmz_objects[objid]->u.missile.damage = MISSILE_DAMAGE;
		dmz_objects[objid]->u.missile.exp_frame = 0;
		dmz_objects[objid]->u.missile.exp_ttl = 0;
		break;
	
	default:
		sprintf(buf, "[common:object:dmz_new_object] asked to create unknown object type %d",
			objtype);
		dmz_logerr(buf);
		break;
	}

	return objid;
}

void dmz_delete_object(WORD16 objid)
{
	dmz_assert(dmz_objects[objid]);
	dmz_free(dmz_objects[objid], sizeof(struct dmz_object));
	dmz_objects[objid] = NULL;
}

/* update status of an object */
void dmz_status_update(WORD16 obj, OBJSTATE status)
{
	dmz_objects[obj]->status = status;
}
