/*	File:	server.c
	Author:	Alex Fedosov
	Date:	Decemer 21, 2000
	Description:	new multiplexing server,
					code for network I/O thread
*/

#include <stdio.h>
#include <stdlib.h>

#include "sys/lock.h"

#include "server.h"

#include "dmzcore.h"
#include "dmz_net.h"
#include "dmzgraph.h"

#define MAXCOLORS 4
#define MAXUPGRADES 4

CLIENT clients[MAXCLIENTS];
int client_ids[MAXPLAYERS];
int num_clients = 0, num_servers = 0;
char buf[256];
WORD16 bootybit;
WORD16 missile, explosion;
WORD16 dead_player;

char player_bitmaps[MAXCOLORS * MAXUPGRADES][64] =
{
	 "data/Robots/RobotA00.bmp",
	 "data/Robots/RobotB00.bmp",
	 "data/Robots/RobotC00.bmp",
	 "data/Robots/RobotD00.bmp",
 	 "data/Robots/RobotA01.bmp",
	 "data/Robots/RobotB01.bmp",
	 "data/Robots/RobotC01.bmp",
	 "data/Robots/RobotD01.bmp",
 	 "data/Robots/RobotA02.bmp",
	 "data/Robots/RobotB02.bmp",
	 "data/Robots/RobotC02.bmp",
	 "data/Robots/RobotD02.bmp",
 	 "data/Robots/RobotA03.bmp",
	 "data/Robots/RobotB03.bmp",
	 "data/Robots/RobotC03.bmp",
	 "data/Robots/RobotD03.bmp",
};

int dmz_server_main(void *arg)
{
	int listenid;
	int active = 1, i;
	struct dmz_message m;
	char buf[256];

	__lock_t *releaseme = (__lock_t *) arg;

	dmz_logstd("[server:server:main] dmz server starting!");

	for (i = 0; i < MAXCLIENTS; i++)
	{
		clients[i].connid = 0;
		clients[i].objid = 0;
		clients[i].ip_addr = NULL;
	}

	/* load player skins before initialization to avoid discrepancies */
	for (i = 0; i < MAXCOLORS * MAXUPGRADES; i++)
	{
		dmz_graph_load_image(player_bitmaps[i], NULL, NULL);
	}
	
	/* these lines are not needed, but necessary for proper display */
	dead_player = dmz_graph_load_image("data/Robots/dead.bmp", NULL, NULL);
	bootybit = dmz_graph_load_image("data/Booty/booty.bmp", NULL, NULL);
	dmz_graph_create_virtual(640, 480);
	dmz_graph_load_image("data/wait.bmp", NULL, NULL);
	
	/* initialize everything */
	i = dmz_data_init();
	if (i != DMZ_SUCCESS)
	{
		sprintf(buf, "[server:server:dmz_server_main] init failed, error: %d", i);
		dmz_die(buf);
		exit(-1);
	}

	dmz_logstd("[server:server:main] network init");
	dmz_net_init();
	listenid = dmz_listen(DMZPORT, MAXPLAYERS);

	for (i = 0; i < MAXPLAYERS; i++)
		client_ids[i] = 0;

	if (releaseme) __mutex_unlock(*releaseme);

	missile = dmz_graph_load_image("data/Special/missile.bmp", NULL, NULL);
	explosion = dmz_graph_load_image("data/Special/explosion.bmp", NULL, NULL);

	dmz_logstd("server:server:main] entering main server loop, listening for connections");
	num_servers++;

	/* run in a loop, wait for connections */
	while (active)
	{
		int accept_ids[2], read_ids[MAXPLAYERS];
		int numA = 1, numR = 0, rv;
		accept_ids[0] = listenid;

		for (i = 0; i < MAXPLAYERS; i++)
			if (client_ids[i] > 0)
				read_ids[numR++] = client_ids[i];

		/* multiplex connections to block until something is available */
		rv = dmz_net_multiplex(read_ids, NULL, accept_ids, &numR, NULL, &numA, NULL);

		/* something bad (tm) happened */
		if (rv < 0)
		{
			active = 0;
			dmz_die("[server:server:main] network error, expect trouble, exiting");
			break;
		}

		/* no data available, retry */
		else if (rv == 0)
			continue;
					
		/* a new connection has been made, and needs to be accepted */
		if (numA > 0)
		{
			int accepted;

			/* we assume it's the connection to listenid, since it's the only listening socket */
			/* add to existing clients */
			for (i = 0; i < MAXPLAYERS; i++)
				if (client_ids[i] == 0)
				{
					accepted = client_ids[i] = dmz_accept(listenid);
					break;
				}

			if (i >= MAXPLAYERS)
				dmz_logstd("[server:server:main] incoming connection but maximum reached");
			else
			{
				sprintf(buf, "[server:server:main] new connection from: %s put in slot %d",
					dmz_net_getpeer(client_ids[i]), i);
				dmz_logstd(buf);
				num_clients++;
				clients[accepted].connid = accepted;
				clients[accepted].ip_addr = dmz_net_getpeer(accepted);
			}
		}

		/* a connection has pending data, read it */
		if (numR > 0)
		{
			for (i = 0; i < numR; i++)
			{
				int rv;
			
				rv = dmz_recv_msg(&m, sizeof(struct dmz_message), read_ids[i]);

				if (rv == sizeof(struct dmz_message))
					dmz_process_msg(&m, read_ids[i]);

				else
				{
					int spill = sizeof(struct dmz_message) - rv;
					sprintf(buf, "[server:server:main] bad receive, returned %d", rv);
					dmz_logerr(buf);
				}

				/* if error occured, drop the client */
				if (rv == DMZ_NET_RECV_ERROR)
					dmz_drop_client(read_ids[i]);
			}
		}
	} /* while */

	return DMZ_SUCCESS;
}

int dmz_process_msg(struct dmz_message *m, int connid)
{
	WORD16 player = 0;

	player = m->objid;

	switch (m->type)
	{
		case BEGIN:
			dmz_process_message_BEGIN(m, connid);
			break;

		case MOVE:
			dmz_process_message_MOVE(m);
			break;

		case CHAT:
			dmz_broadcast(m);
			break;

		case ATTACK:
			if (dmz_objects[m->objid]->status == DEAD) break;

			if (m->u.attack.method == PROJECTILE)
			{
				MissileAttack(m->objid, m->u.attack.direction);
				dmz_broadcast(m);
			}

			else if (m->u.attack.method == HTH)
			{
				HTHAttack(m->objid, m->u.attack.direction);
				dmz_broadcast(m);
			}

			else
			{
				sprintf(buf,
				"[server:server:process_msg] unknown attack type %d from %d",
				m->u.attack.method, player);
			}

			break;

//		case OBJECT:	/* object transfer -- player wants an object */
//			/* copy the object into the message and send it */
//			memcpy(&(m->u.object.obj), dmz_objects[player], sizeof(struct dmz_object));
//			rv = dmz_send_msg(m, sizeof(struct dmz_message), connid);
//			if (rv == DMZ_NET_SEND_ERROR)
//				dmz_drop_client(connid);
//			break;

		case END:		/* player is quitting, put him to sleep */
			sprintf(buf, "END message from connid %d", connid);
			dmz_logstd(buf);

			if (dmz_objects[player]->status != DEAD)
				dmz_status_update(player, SLEEPING);

			dmz_drop_client(connid);

			/* broadcast the message to everyone */
			dmz_broadcast(m);
			break;

		default:		/* unknown message type */
			sprintf(buf, "[server:server:process_msg] unknown message type %d from %d",
				m->type, player);
			dmz_logerr(buf);

			return DMZ_BAD_MESSAGE;
			break;
	}

	return DMZ_SUCCESS;
}

void dmz_drop_client(int connid)
{
	int i;

	/* TODO set the appropriate client id to zero */
	dmz_disconnect(connid);

	num_clients--;

	for (i = 0; i < MAXPLAYERS; i++)
		if (client_ids[i] == connid)
		{
			client_ids[i] = 0;
			break;
		}

	clients[connid].connid = 0;
	clients[connid].objid = 0;
	clients[connid].ip_addr = NULL;
}

int dmz_broadcast(struct dmz_message *m)
{
	int i, rv;

	for (i = 0; i < MAXPLAYERS; i++)
	{
		if (client_ids[i] != 0)
		{
			rv = dmz_send_msg(m, sizeof(struct dmz_message), client_ids[i]);

			if (rv == DMZ_NET_SEND_ERROR)
				dmz_drop_client(client_ids[i]);
		}
	}

	return DMZ_SUCCESS;
}

int dmz_process_message_BEGIN(struct dmz_message *m, int connid)
{
	WORD32 x, y; 
	WORD16 player = m->objid;
	int player_existed = 0;		/* flag to indicate returning players */
	int i, rv;

	/* select random coordinates for respawn */
	/* TODO collision detection here! */
	do
	{
		x = (rand() % map_width) * PIXELS_PER_CELL;
		y = (rand() % map_height) * PIXELS_PER_CELL;
	}
	while (!CellPassable(x, y));

	sprintf(buf, "begin: selected (%d, %d) as starting position", x, y);
	dmz_logstd(buf);

	/* assignment of an object id, this gets pretty complicated */

	/* if asked to create a new player or requested player does not exist */
	if (player == 0)
	{
		player = dmz_new_object(0, PLAYER, x, y);
		sprintf(buf, "[server:server:BEGIN] created new object %d for client %d", player, connid);
		dmz_logstd(buf);
	}

	/* player requested to use existing object */
	else if (player > 0 && player < MAXOBJECTS)
	{
		/* asked for an object that does not exist, create new one */
		if (dmz_objects[player] == NULL)
		{
			player = dmz_new_object(0, PLAYER, x, y);
			sprintf(buf, "[server:server:BEGIN] asked for nonexisting object created %d client %d", player, connid);
			dmz_logstd(buf);
		}

		else
		{
			/* dead player respawning, reset properties */
			if (dmz_objects[player]->status == DEAD)
			{
				sprintf(buf, "[server:server:BEGIN] respawning dead player %d client %d", player, connid);
				dmz_logstd(buf);
				player = dmz_new_object(player, PLAYER, x, y);
			}

			/* object exists, make sure it's hybernating and 
			   not used by anybody else. also check for name match */
			else if (dmz_objects[player]->status == SLEEPING &&
				strncmp(m->u.begin.name, dmz_objects[player]->name, NAMELEN) == 0)
			{
				/* do nothing, really, just use that id */
				sprintf(buf, "[server:server:BEGIN] using existing object %d client %d", player, connid);
				dmz_logstd(buf);
				player_existed = 1;
			}

			else	/* object in use or name mismatch, get new one */
			{
				player = dmz_new_object(0, PLAYER, x, y);
				sprintf(buf, "[server:server:BEGIN] asked for object in use created %d client %d", player, connid);
				dmz_logstd(buf);
			}
		}
	}

	/* id out of range */
	else
		return DMZ_BAD_MESSAGE;


	clients[connid].objid = player;

	/* update status to indicate that player is alive */
	dmz_status_update(player, IDLE);

	if (!player_existed)
	{
		strtoupr(m->u.begin.name);
		memcpy(dmz_objects[player]->name, m->u.begin.name, NAMELEN);
				
		/* assign new position */
		dmz_objects[player]->x = x;
		dmz_objects[player]->y = y;
		dmz_objects[player]->u.player.direction = DEFAULT_DIR;
		dmz_objects[player]->bitidx = m->u.begin.bitmap;
		dmz_objects[player]->deadbit = dead_player;
	}

	/* update the structure and send it back */
	m->objid = player;
	m->u.begin.init_x = dmz_objects[player]->x;
	m->u.begin.init_y = dmz_objects[player]->y;
	m->u.begin.init_dir = dmz_objects[player]->u.player.direction;
	m->u.begin.bitmap = dmz_objects[player]->bitidx;

	/* broadcast message to all players */
	dmz_broadcast(m);

	/* initial update of player's stats */
	memset(m, 0, sizeof(struct dmz_message));
	m->objid = player;
	m->type = UPDATE;
	m->u.update.life = dmz_objects[player]->u.player.life;
	m->u.update.score = dmz_objects[player]->score;
	m->u.update.ammo = dmz_objects[player]->u.player.ammo;
	m->u.update.frags = -1;
	m->u.update.status = -1;
	dmz_broadcast(m);

	/* send over all the current objects */
	for (i = 1; i < MAXOBJECTS; i++)
	{
		struct dmz_message m_obj;

		if (dmz_objects[i] == NULL || dmz_objects[i]->objid == player) continue;

		dmz_assert(dmz_objects[i]->objid);

		memset(&m_obj, 0, sizeof(struct dmz_message));
		m_obj.type = OBJECT;
		m_obj.objid = dmz_objects[i]->objid;
		memcpy(&(m_obj.u.object.obj), dmz_objects[i], sizeof(struct dmz_object));

		sprintf(buf, "sending object %d -- %s bitmap %d", i, dmz_objects[i]->name, dmz_objects[i]->bitidx);
		dmz_logstd(buf);

		rv = dmz_send_msg(&m_obj, sizeof(struct dmz_message), connid);

		/* if error occurred, drop the client */
		if (rv == DMZ_NET_SEND_ERROR)
		{
			dmz_drop_client(connid);
			break;
		}
	}

	return DMZ_SUCCESS;
}

int dmz_process_message_MOVE(struct dmz_message *m)
{
	/* assuming collision detection is done on the client side,
	   we don't verify the validity of the move (yet) and simply pass it on. */

	CELL cell;
	WORD16 colid;
	WORD16 player = m->objid;
	WORD32 new_x, new_y;
	short pass, spd, dam;

	/* you're dead, dammit! go away. */
	if (dmz_objects[m->objid]->status == DEAD) return DMZ_INVALID_MOVE;

	/* update player's position */
	new_x = m->u.move.new_x;
	new_y = m->u.move.new_y;

	/* check if out of bounds */
	if (new_x > map_width * PIXELS_PER_CELL ||
		new_y > map_height * PIXELS_PER_CELL)
	{
		return DMZ_INVALID_MOVE;
	}

	if (!ValidMove(m->u.move.direction, player))
	{
		sprintf(buf, "player %d invalid move to (%d, %d)", player, new_x, new_y);
		dmz_logerr(buf);
		return DMZ_INVALID_MOVE;
	}

	colid = DetectCollision(player);
	if (colid > 0)
	{
		/* collision with items is allowed, we pick it up */
		if (dmz_objects[colid]->objtype == ITEM)
		{
			PickupItem(player, colid);
		}
		
		/* we're allowed to run over small terrain objects */
		else if (dmz_objects[colid]->objtype == TERRAIN &&
			dmz_objects[colid]->width < 64)
		{
			struct dmz_message m_dead;
			memset(&m_dead, 0, sizeof(struct dmz_message));

			/* kill that thing */
			dmz_status_update(colid, DEAD);

			/* ... and tell everyone about it */
			m_dead.type = DEATH;
			m_dead.objid = colid;
			dmz_broadcast(&m_dead);
		}

		else
		{
			sprintf(buf, "player %d collided at (%d, %d)", player, new_x, new_y);
			dmz_logerr(buf);
			/* TODO allow collisions for now */
			//return DMZ_INVALID_MOVE;
		}
	}

	/* apply damage and speed change, if applicable */
	cell = dmz_world_map[new_x / PIXELS_PER_CELL][new_y / PIXELS_PER_CELL];
	dmz_map_get_prop(cell, &pass, &dam, &spd);

	dmz_objects[player]->velocity = DEFAULT_VELOCITY + spd;

	/* everything is cool, update the object */
	dmz_objects[player]->x = new_x;
	dmz_objects[player]->y = new_y;

	/* update player's direction */
	dmz_objects[player]->u.player.direction = m->u.move.direction;

	if (dam != 0)
	{
		struct dmz_message m_upd;
		memset(&m_upd, 0, sizeof(struct dmz_message));

		ApplyDamage(player, dam);

		/* update of player's life */
		m_upd.objid = player;
		m_upd.type = UPDATE;
		m_upd.u.update.life = dmz_objects[player]->u.player.life;
		m_upd.u.update.score = -1;
		m_upd.u.update.ammo = -1;
		m_upd.u.update.status = -1;
		m_upd.u.update.frags = -1;

		dmz_broadcast(&m_upd);
	}

	/* broadcast move message to everyone */
	dmz_broadcast(m);

	return DMZ_SUCCESS;
}

void dump_server_stats(void)
{
	int numobjs = 0;
	int numcells = 0;
	int numplayers = 0;
	int numobjects = 0;
	int i;

	sprintf(buf, "[server:server:dump_server_stats] dmz server version 0x%4X", DMZ_SERVER_VERSION);
	dmz_logstd(buf);

	for (i = 0; i < MAXOBJECTS; i++)
	{
		if (dmz_objects[i] != NULL)
		{
			if (dmz_objects[i]->objtype == PLAYER)
				numplayers++;

			numobjs++;
		}
	}

	for (i = 0; i < MAXCELLS; i++)
	{
		if (dmz_map_cells[i] != NULL) numcells++;
	}

	sprintf(buf, "[server:server:dump_server_stats] server info:\
		\nobject size: %d\tmessage size: %d\tmap cell size: %d\
		\nworld map: %d\tcell array: %d\tobject array: %d\
		\n%d players, %d objects, %d map cells",
		sizeof(struct dmz_object), sizeof(struct dmz_message), sizeof(struct dmz_map_cell),
		sizeof(dmz_world_map), sizeof(dmz_map_cells), sizeof(dmz_objects),
		numplayers, numobjects, numcells);
	dmz_logstd(buf);
}

int main(void)
{
	dmz_server_main(NULL);
	return 0;
}

int PickupItem(WORD16 objid, WORD16 item)
{
	struct dmz_message m, m2;
	memset(&m, 0, sizeof(struct dmz_message));
	memset(&m2, 0, sizeof(struct dmz_message));
	
	/* update of player's stats */
	m.type = UPDATE;
	m.objid = objid;
	m.u.update.life = -1;
	m.u.update.score = -1;
	m.u.update.status = -1;
	m.u.update.ammo = -1;
	m.u.update.frags = -1;

	/* update of player's stats */
	m2.type = UPDATE;
	m2.objid = item;
	m2.u.update.life = -1;
	m2.u.update.score = -1;
	m2.u.update.status = INVISIBLE;
	m2.u.update.ammo = -1;
	m2.u.update.frags = -1;

	if (dmz_objects[item]->u.item.kind > 0)
	{
		switch (dmz_objects[item]->u.item.kind)
		{
			case ARMORUPGRADE:
				sprintf(buf,"player life is %d",dmz_objects[objid]->u.player.life);
				dmz_logstd(buf);

				if (dmz_objects[objid]->u.player.life < DEFAULT_LIFE)
				{
					dmz_objects[objid]->u.player.life += dmz_objects[item]->u.item.item_arm;
					if (dmz_objects[objid]->u.player.life > DEFAULT_LIFE)
						dmz_objects[objid]->u.player.life = DEFAULT_LIFE;
				}

				m.u.update.life = dmz_objects[objid]->u.player.life;
				
				sprintf(buf,"now player life is %d",dmz_objects[objid]->u.player.life);
				dmz_logstd(buf);

				dmz_status_update(item, INVISIBLE);
				dmz_objects[item]->tstamp = time(NULL);
				//dmz_delete_object(item);

				dmz_broadcast(&m2);
				break;

			case DAMAGEUPGRADE:
				break;
			
			case AMMOUPGRADE:
				if (dmz_objects[objid]->u.player.ammo < DEFAULT_AMMO)
				{
					dmz_objects[objid]->u.player.ammo += dmz_objects[item]->u.item.item_ammo;
					if (dmz_objects[objid]->u.player.ammo > DEFAULT_AMMO)
						dmz_objects[objid]->u.player.ammo = DEFAULT_AMMO;
				}

				m.u.update.ammo = dmz_objects[objid]->u.player.ammo;

				dmz_status_update(item, INVISIBLE);
				dmz_objects[item]->tstamp = time(NULL);
				//dmz_delete_object(item);
				
				dmz_broadcast(&m2);
				break;

			case BOOTY:
				if (dmz_objects[objid]->u.player.ammo < DEFAULT_AMMO)
				{
					dmz_objects[objid]->u.player.ammo += dmz_objects[item]->u.item.item_ammo;
					if (dmz_objects[objid]->u.player.ammo > DEFAULT_AMMO)
						dmz_objects[objid]->u.player.ammo = DEFAULT_AMMO;
				}

				if (dmz_objects[objid]->u.player.life < DEFAULT_LIFE)
				{
					dmz_objects[objid]->u.player.life += dmz_objects[item]->u.item.item_arm;
					if (dmz_objects[objid]->u.player.life > DEFAULT_LIFE)
						dmz_objects[objid]->u.player.life = DEFAULT_LIFE;
				}

				m.u.update.life = dmz_objects[objid]->u.player.life;
				m.u.update.ammo = dmz_objects[objid]->u.player.ammo;

				//dmz_status_update(item, INVISIBLE);
				//dmz_objects[item]->tstamp = time(NULL);
				/* delete the boots, we don't want them to regen */
				dmz_delete_object(item);
				break;

			default:
				break;
		}
	}

	dmz_broadcast(&m);
	return DMZ_SUCCESS;
}

void Regenerate(WORD16 objid)
{
	struct dmz_message m;
	memset(&m, 0, sizeof(struct dmz_message));

	m.objid = objid;
	m.type = UPDATE;
	m.u.update.life = -1;
	m.u.update.score = -1;
	m.u.update.frags = -1;
	m.u.update.ammo = -1;
	m.u.update.status = -1;

	if (dmz_objects[objid]->status != INVISIBLE &&
		dmz_objects[objid]->status != DEAD)
		return;

	if (dmz_objects[objid]->objtype == MONSTER)
	{
		sprintf(buf, "attempting to regen monster %d", objid);
		dmz_logstd(buf);

		if (time(NULL) > (dmz_objects[objid]->tstamp + MONSTER_REGEN))
		{
			WORD32 x, y;

			/* select random coordinates for respawn */
			/* TODO collision detection here! */
			do
			{
				x = (rand() % map_width) * PIXELS_PER_CELL;
				y = (rand() % map_height) * PIXELS_PER_CELL;
			}
			while (!CellPassable(x, y));

			dmz_status_update(objid, IDLE);
			m.u.update.status = IDLE;
			m.u.update.life = DEFAULT_LIFE;
			dmz_broadcast(&m);

			dmz_objects[objid]->u.player.life = DEFAULT_LIFE;

			/* update position */
			dmz_objects[objid]->x = x;
			dmz_objects[objid]->y = y;

			sprintf(buf, "monster regenerated %s tstamp was %d", dmz_objects[objid]->name, dmz_objects[objid]->tstamp);
			dmz_logstd(buf);
			sprintf(buf, "time difference %d", time(NULL) - dmz_objects[objid]->tstamp);
			dmz_logstd(buf);
		}
	}
		
	else if (dmz_objects[objid]->objtype == ITEM)
	{
		sprintf(buf, "attempting to regen item %d", objid);
		dmz_logstd(buf);

		if (time(NULL) > (dmz_objects[objid]->tstamp + ITEM_REGEN))
		{
			dmz_status_update(objid, IDLE);
			m.u.update.status = IDLE;
			dmz_broadcast(&m);

			sprintf(buf, "item regenerated %s", dmz_objects[objid]->name);
			dmz_logstd(buf);
		}
	}
}

void DrawProgressIndicator(int percent, int flag)
{
}
